use strict;
use HTTP::Daemon;
use HTTP::Status;
use HTTP::Headers;
use HTTP::Response;
use GD;
use GD::Text;
use Win32::SerialPort qw(:STAT);
use Win32::Registry;
use Time::HiRes qw/sleep/;

require URI;

my $Registry;
$::HKEY_LOCAL_MACHINE->Open("HARDWARE\\DEVICEMAP\\SERIALCOMM", $Registry) or die "Can't open data: $^E";
my @PortNames = ();
my $Port;

$| = 1;

my $shell = $ENV{ComSpec};
system "$shell /c start /max http://localhost:3456";

my $img = GD::Image->new(128, 96);
$img->colorAllocate($_ * 16, $_ * 16, $_ * 16) foreach (0 .. 15);

my $connecting = GD::Image->new(384, 288);
$connecting->colorAllocate(255, 255, 255);
$connecting->colorAllocate(0, 0, 0);
$connecting->string(gdLargeFont, 100, 140, "Connecting to PropCAM...", 1); 
$connecting = $connecting->gif;

my $notfound = GD::Image->new(384, 288);
$notfound->colorAllocate(255, 0, 0);
$notfound->colorAllocate(255, 255, 255);
$notfound->string(gdLargeFont, 130, 130, "PropCAM not found.", 1); 
$notfound->string(gdLargeFont, 140, 145, "Server offline.", 1); 
$notfound = $notfound->gif;

my $offline = GD::Image->new(384, 288);
$offline->colorAllocate(0, 0, 0);
$offline->colorAllocate(255, 255, 255);
$offline->string(gdLargeFont, 140, 140, "Server offline.", 1); 
$offline = $offline->gif;

my $daemon = HTTP::Daemon->new(LocalPort => 3456) || die;
print "PropCAM Image Server\n\n";
while (my ($connection, $peer) = $daemon->accept) {
	my $client = join('.', map {ord($_)} split(//, substr($peer, 4, 4)));
	print "Connection from $client:\n";
	while (my $req = $connection->get_request) {
		$connection->send_error(RC_FORBIDDEN) unless $client eq '127.0.0.1';
    if ($req->method eq 'GET') {
    	my $path = $req->url->path;
    	print "  $path\n";
    	
	    if ($path eq '/') {
    		respond($connection, 'text/html', main_page());
    		
    	} elsif ($path =~ m/connecting/) {
    		respond($connection, 'image/gif', $connecting);
    		open_port();
    		
    	} elsif ($path =~ m/image.gif/) {
    		if (defined $Port) {
    			$Port->input;
    			$Port->write("S");
    			my ($cnt, $duo);
	    		foreach my $y (0 .. 95) {
	    			foreach my $x (0 .. 63) {	    				
	    				do {
	    					($cnt, $duo) = $Port->read(1);
	    				} until $cnt;
	    				$duo = ord($duo);
	    				$img->setPixel($x * 2, $y, $duo & 0x0f);
	    				$img->setPixel($x * 2 + 1, $y, $duo >> 4);
						}
					}
	    		respond($connection, 'image/gif', $img->gif)
	    	} else {
					$connection->send_error(RC_NOT_FOUND)
				}	    		
					
			} elsif ($path =~ m/notfound/) {
  			respond($connection, 'image/gif', $notfound);
  			exit
    		
    	} elsif ($path =~ m/exit/) {
    		respond($connection, 'image/gif', $offline);
    		exit
    		
			} else {
				$connection->send_error(RC_NOT_FOUND)
			}
		} else {
			$connection->send_error(RC_FORBIDDEN)
		}
	}
	$connection->close;
	undef($connection);
}

sub respond {
	my $con = shift;
	my $typ = shift;
	my $txt = shift;
	my $header = HTTP::Headers->new;
	$header->date( time );
	$header->server("$0");
	$header->content_type($typ);
	my $response = HTTP::Response->new(RC_OK);
	$response->content($txt);
	$response->header($header);
	$con->send_response( $response )
}

sub open_port {
	return if defined $Port;
	my %coms;
	$Registry->GetValues(\%coms);
	@PortNames = grep {m/^COM\d+$/} (map {$coms{$_}->[2]} keys %coms);
	foreach my $dev ((@PortNames) x 2, undef) {
		unless (defined $dev) {
			undef $Port;
			return
		}
		unless ($Port = new Win32::SerialPort("\\\\.\\$dev")
			and $Port->baudrate(115200)
			and $Port->parity('none')
			and $Port->databits(8)
			and $Port->stopbits(2)
			and $Port->buffers(10000, 100)
			and $Port->binary(1)
			and $Port->write_settings)
		{
			next
		}
		$Port->write("I");
		sleep 1;
		my (undef, $n, undef, $err) = ($Port->status);
		if ($err) {
			$Port->reset_error;
		}
		if ($n) {
			my $stream = $Port->input;
			last if $stream =~ m/PropCAM/
		}
		$Port->close;
	}
	return 
}

sub main_page {
	return <<'EOF';
<html xmlns="http://www.w3.org/1999/xhtml" lang="en">
	<head>
		<meta http-equiv="Content-Type" content="text/html; charset=ISO-8859-1" />
	</head>
	<body bgcolor="#ffffd0">
		<script>
			var STARTING = 0;
			var SNAPSHOT = 1;
			var VIDEO = 2;
			var DONE = 3;
			var image;
			var source = "/image.gif";
			var mode = STARTING;
			
			function snapshot() {
				if (mode == DONE || mode == STARTING) {return;}
				mode = SNAPSHOT;
				new_image();
			}
		
			function continuous() {
				if (mode == DONE || mode == STARTING) {return;}
				mode = VIDEO;
				new_image();
			}

			function quit() {
				if (mode == DONE || mode == STARTING) {return;}
				if (confirm("Stop server?")) {
					mode = DONE;
					source = "/exit.gif";
					new_image();
				}				
			}
			
			function refresh() {
				if (mode == DONE) {return;}
				if (mode == STARTING) {
					image = document.getElementById("camera_image");
					mode = SNAPSHOT;
					new_image();
				}
				if (mode == VIDEO) {
					new_image();
				}
			}
			
			function error() {
				mode = DONE;
				image.src = "notfound";
			}
			
			function new_image() {
				image.src = source + "?" + (new Date()).getTime();
			}				
		</script>
		<div align="center">
			<font size="5", face="Arial,sans-serif", color="#000040"><b>PropCAM Monitor</b></font><p/>
			<img id = "camera_image" src="/connecting.gif" width="384" height="288" onLoad = "refresh()" onError = "error()"></img>
			<p/>
			<button onclick="snapshot()">Snap</button>&nbsp;
			<button onclick="continuous()">Video</button>&nbsp;
			<button onclick="quit()">Quit</button>
		</div>
	</body>
</html>
EOF
}
